@@ -18,6 +18,7 @@ from .. import rendering
 from .. import resources
 from .. import transformations
 
+from ..util import unique_name
 from ..caching import hash_fast
 from ..constants import log, tol
 
@@ -61,6 +62,9 @@ _GL_POINTS = 0
 _GL_TRIANGLES = 4
 _GL_STRIP = 5
 
+_EYE = np.eye(4)
+_EYE.flags.writeable = False
+
 # specify dtypes with forced little endian
 float32 = np.dtype("<f4")
 uint32 = np.dtype("<u4")
@@ -71,7 +75,8 @@ def export_gltf(scene,
                 include_normals=None,
                 merge_buffers=False,
                 unitize_normals=False,
-                tree_postprocessor=None):
+                tree_postprocessor=None,
+                embed_buffers=False):
     """
     Export a scene object as a GLTF directory.
 
@@ -90,6 +95,8 @@ def export_gltf(scene,
       If passed will use to write each file.
     tree_postprocesser : None or callable
       Run this on the header tree before exiting.
+    embed_buffers : bool
+      Embed the buffer into JSON file as a base64 string in the URI
 
     Returns
     ----------
@@ -114,13 +121,18 @@ def export_gltf(scene,
     # store files as {name : data}
     files = {}
 
+    base64_buffer_format = "data:application/octet-stream;base64,{}"
     if merge_buffers:
         views = _build_views(buffer_items)
-        buffer_name = "gltf_buffer.bin"
         buffer_data = bytes().join(buffer_items.values())
+        if embed_buffers:
+            buffer_name = base64_buffer_format.format(
+                base64.b64encode(buffer_data).decode())
+        else:
+            buffer_name = "gltf_buffer.bin"
+            files[buffer_name] = buffer_data
         buffers = [{"uri": buffer_name,
                     "byteLength": len(buffer_data)}]
-        files[buffer_name] = buffer_data
     else:
         # make one buffer per buffer_items
         buffers = [None] * len(buffer_items)
@@ -131,10 +143,14 @@ def export_gltf(scene,
             views[i] = {"buffer": i,
                         "byteOffset": 0,
                         "byteLength": len(item)}
-            buffer_name = "gltf_buffer_{}.bin".format(i)
+            if embed_buffers:
+                buffer_name = base64_buffer_format.format(
+                    base64.b64encode(item).decode())
+            else:
+                buffer_name = "gltf_buffer_{}.bin".format(i)
+                files[buffer_name] = item
             buffers[i] = {"uri": buffer_name,
                           "byteLength": len(item)}
-            files[buffer_name] = item
 
     if len(buffers) > 0:
         tree["buffers"] = buffers
@@ -153,7 +169,8 @@ def export_glb(
         scene,
         include_normals=None,
         unitize_normals=False,
-        tree_postprocessor=None):
+        tree_postprocessor=None,
+        buffer_postprocessor=None):
     """
     Export a scene as a binary GLTF (GLB) file.
 
@@ -183,7 +200,7 @@ def export_glb(
     tree, buffer_items = _create_gltf_structure(
         scene=scene,
         unitize_normals=unitize_normals,
-        include_normals=include_normals)
+        include_normals=include_normals, buffer_postprocessor=buffer_postprocessor)
 
     # allow custom postprocessing
     if tree_postprocessor is not None:
@@ -279,9 +296,23 @@ def load_gltf(file_obj=None,
         # old versions of python/json need strings
         tree = json.loads(util.decode_text(data))
 
+    # gltf 1.0 is a totally different format
+    # that wasn't widely deployed before they fixed it
+    version = tree.get('asset', {}).get('version', '2.0')
+    if isinstance(version, str):
+        # parse semver like '1.0.1' into just a major integer
+        major = int(version.split('.', 1)[0])
+    else:
+        major = int(float(version))
+
+    if major < 2:
+        raise NotImplementedError(
+            'only GLTF 2 is supported not `{}`'.format(
+                version))
+
     # use the URI and resolver to get data from file names
     buffers = [_uri_to_bytes(uri=b['uri'], resolver=resolver)
-               for b in tree['buffers']]
+               for b in tree.get('buffers', [])]
 
     # turn the layout header and data into kwargs
     # that can be used to instantiate a trimesh.Scene object
@@ -312,7 +343,8 @@ def load_glb(file_obj,
     resolver : trimesh.visual.Resolver
       Object which can be used to load other files by name
     merge_primitives : bool
-      If True, each GLTF 'mesh' will correspond to a single Trimesh object
+      If True, each GLTF 'mesh' will correspond to a
+      single Trimesh object.
 
     Returns
     ------------
@@ -327,10 +359,15 @@ def load_glb(file_obj,
     head_data = file_obj.read(20)
     head = np.frombuffer(head_data, dtype="<u4")
 
-    # check to make sure first index is gltf
-    # and second is 2, for GLTF 2.0
-    if head[0] != _magic["gltf"] or head[1] != 2:
-        raise ValueError("file is not GLTF 2.0")
+    # check to make sure first index is gltf magic header
+    if head[0] != _magic["gltf"]:
+        raise ValueError('incorrect header on GLB file')
+
+    # and second value is version: should be 2 for GLTF 2.0
+    if head[1] != 2:
+        raise NotImplementedError(
+            'only GLTF 2 is supported not `{}`'.format(
+                head[1]))
 
     # overall file length
     # first chunk length
@@ -378,6 +415,7 @@ def load_glb(file_obj,
                            ignore_broken=ignore_broken,
                            merge_primitives=merge_primitives,
                            mesh_kwargs=mesh_kwargs)
+
     return kwargs
 
 
@@ -497,7 +535,7 @@ def _data_append(acc, buff, blob, data):
 
     if kind == 'SCALAR':
         # is probably (n, 1)
-        blob['count'] = int(np.product(data.shape))
+        blob['count'] = int(np.prod(data.shape))
         blob['max'] = np.array(
             [data.max()], dtype=dtype).tolist()
         blob['min'] = np.array(
@@ -564,7 +602,8 @@ def _jsonify(blob):
 def _create_gltf_structure(scene,
                            include_normals=None,
                            include_metadata=True,
-                           unitize_normals=None,):
+                           unitize_normals=None,
+                           buffer_postprocessor=None):
     """
     Generate a GLTF header.
 
@@ -608,8 +647,12 @@ def _create_gltf_structure(scene,
             # fail here if data isn't json compatible
             # only export the extras if there is something there
             tree['scenes'][0]['extras'] = _jsonify(scene.metadata)
+            extensions = tree['scenes'][0]['extras'].pop(
+                'gltf_extensions', None)
+            if isinstance(extensions, dict):
+                tree['scenes'][0]['extensions'] = extensions
         except BaseException:
-            log.warning(
+            log.debug(
                 'failed to export scene metadata!', exc_info=True)
 
     # store materials as {hash : index} to avoid duplicates
@@ -658,6 +701,25 @@ def _create_gltf_structure(scene,
         scene=scene, mesh_index=mesh_index)
     tree.update(nodes)
 
+    extensions_used = set()
+    # Add any scene extensions used
+    if 'extensions' in tree['scenes'][0]:
+        extensions_used = extensions_used.union(
+            set(tree['scenes'][0]['extensions'].keys()))
+    # Add any mesh extensions used
+    for mesh in tree['meshes']:
+        if 'extensions' in mesh:
+            extensions_used = extensions_used.union(
+                set(mesh['extensions'].keys()))
+    # Add any extensions already in the tree (e.g. node extensions)
+    if 'extensionsUsed' in tree:
+        extensions_used = extensions_used.union(set(tree['extensionsUsed']))
+    if len(extensions_used) > 0:
+        tree['extensionsUsed'] = list(extensions_used)
+
+    if buffer_postprocessor is not None:
+        buffer_postprocessor(buffer_items, tree)
+
     # convert accessors back to a flat list
     tree['accessors'] = list(tree['accessors'].values())
 
@@ -702,7 +764,7 @@ def _append_mesh(mesh,
     """
     # return early from empty meshes to avoid crashing later
     if len(mesh.faces) == 0 or len(mesh.vertices) == 0:
-        log.warning('skipping empty mesh!')
+        log.debug('skipping empty mesh!')
         return
     # convert mesh data to the correct dtypes
     # faces: 5125 is an unsigned 32 bit integer
@@ -739,11 +801,16 @@ def _append_mesh(mesh,
         # skip jsonify any metadata, skipping internal keys
         current['extras'] = _jsonify(mesh.metadata)
 
+        # extract extensions if any
+        extensions = current['extras'].pop('gltf_extensions', None)
+        if isinstance(extensions, dict):
+            current['extensions'] = extensions
+
         if mesh.units not in [None, 'm', 'meters', 'meter']:
             current["extras"]["units"] = str(mesh.units)
     except BaseException:
-        log.warning('metadata not serializable, dropping!',
-                    exc_info=True)
+        log.debug('metadata not serializable, dropping!',
+                  exc_info=True)
 
     # check to see if we have vertex or face colors
     # or if a TextureVisual has colors included as an attribute
@@ -772,7 +839,7 @@ def _append_mesh(mesh,
 
     if hasattr(mesh.visual, 'material'):
         # append the material and then set from returned index
-        current["primitives"][0]["material"] = _append_material(
+        current_material = _append_material(
             mat=mesh.visual.material,
             tree=tree,
             buffer_items=buffer_items,
@@ -795,19 +862,23 @@ def _append_mesh(mesh,
                                         "byteOffset": 0},
                                   data=uv.astype(float32))
             # add the reference for UV coordinates
-            current["primitives"][0]["attributes"][
-                "TEXCOORD_0"] = acc_uv
+            current["primitives"][0]["attributes"]["TEXCOORD_0"] = acc_uv
+            # only reference the material if we had UV coordinates
+        current["primitives"][0]["material"] = current_material
 
     if (include_normals or
         (include_normals is None and
          'vertex_normals' in mesh._cache.cache)):
         # store vertex normals if requested
-        #
-        normals = mesh.vertex_normals.copy()
         if unitize_normals:
+            normals = mesh.vertex_normals.copy()
             norms = np.linalg.norm(normals, axis=1)
             if not util.allclose(norms, 1.0, atol=1e-4):
                 normals /= norms.reshape((-1, 1))
+        else:
+            # we don't have to copy them since
+            # they aren't being altered
+            normals = mesh.vertex_normals
 
         acc_norm = _data_append(
             acc=tree['accessors'],
@@ -828,13 +899,21 @@ def _append_mesh(mesh,
         # prefixed with an underscore
         if not key.startswith("_"):
             key = "_" + key
+
+        # GLTF has no floating point type larger than 32 bits so clip
+        # any float64 or larger to float32
+        if attrib.dtype.kind == 'f' and attrib.dtype.itemsize > 4:
+            data = attrib.astype(np.float32)
+        else:
+            data = attrib
+
         # store custom vertex attributes
         current["primitives"][0][
             "attributes"][key] = _data_append(
                 acc=tree['accessors'],
                 buff=buffer_items,
-                blob=_build_accessor(attrib),
-                data=attrib)
+                blob=_build_accessor(data),
+                data=data)
 
     tree["meshes"].append(current)
 
@@ -887,7 +966,8 @@ def _build_accessor(array):
     if len(shape) == 2:
         vec_length = shape[1]
         if vec_length > 4:
-            raise ValueError("The GLTF spec does not support vectors larger than 4")
+            raise ValueError(
+                "The GLTF spec does not support vectors larger than 4")
         if vec_length > 1:
             data_type = "VEC%d" % vec_length
         else:
@@ -899,7 +979,7 @@ def _build_accessor(array):
         data_type = "MAT%d" % shape[2]
 
     # get the array data type as a str stripping off endian
-    lookup = array.dtype.str[-2:]
+    lookup = array.dtype.str.lstrip('<>')
 
     if lookup == 'u4':
         # spec: UNSIGNED_INT is only allowed when the accessor
@@ -1012,8 +1092,8 @@ def _append_path(path, name, tree, buffer_items):
     try:
         current["extras"] = _jsonify(path.metadata)
     except BaseException:
-        log.warning('failed to serialize metadata, dropping!',
-                    exc_info=True)
+        log.debug('failed to serialize metadata, dropping!',
+                  exc_info=True)
 
     if path.colors is not None:
         acc_color = _data_append(acc=tree['accessors'],
@@ -1130,6 +1210,40 @@ def specular_to_pbr(
     return result
 
 
+def _parse_textures(header, views, resolver=None):
+    try:
+        import PIL.Image
+    except ImportError:
+        log.debug("unable to load textures without pillow!")
+        return None
+
+    # load any images
+    images = None
+    if "images" in header:
+        # images are referenced by index
+        images = [None] * len(header["images"])
+        # loop through images
+        for i, img in enumerate(header["images"]):
+            # get the bytes representing an image
+            if 'bufferView' in img:
+                blob = views[img["bufferView"]]
+            elif 'uri' in img:
+                # will get bytes from filesystem or base64 URI
+                blob = _uri_to_bytes(uri=img['uri'], resolver=resolver)
+            else:
+                log.debug('unable to load image from: {}'.format(
+                    img.keys()))
+                continue
+            # i.e. 'image/jpeg'
+            # mime = img['mimeType']
+            try:
+                # load the buffer into a PIL image
+                images[i] = PIL.Image.open(util.wrap_as_stream(blob))
+            except BaseException:
+                log.error("failed to load image!", exc_info=True)
+    return images
+
+
 def _parse_materials(header, views, resolver=None):
     """
     Convert materials and images stored in a GLTF header
@@ -1147,36 +1261,7 @@ def _parse_materials(header, views, resolver=None):
     materials : list
       List of trimesh.visual.texture.Material objects
     """
-    try:
-        import PIL.Image
-    except ImportError:
-        log.warning("unable to load textures without pillow!")
-        return None
-
-    # load any images
-    images = None
-    if "images" in header:
-        # images are referenced by index
-        images = [None] * len(header["images"])
-        # loop through images
-        for i, img in enumerate(header["images"]):
-            # get the bytes representing an image
-            if 'bufferView' in img:
-                blob = views[img["bufferView"]]
-            elif 'uri' in img:
-                # will get bytes from filesystem or base64 URI
-                blob = _uri_to_bytes(uri=img['uri'], resolver=resolver)
-            else:
-                log.warning('unable to load image from: {}'.format(
-                    img.keys()))
-                continue
-            # i.e. 'image/jpeg'
-            # mime = img['mimeType']
-            try:
-                # load the buffer into a PIL image
-                images[i] = PIL.Image.open(util.wrap_as_stream(blob))
-            except BaseException:
-                log.error("failed to load image!", exc_info=True)
+    images = _parse_textures(header, views, resolver)
 
     # store materials which reference images
     materials = []
@@ -1201,58 +1286,19 @@ def _parse_materials(header, views, resolver=None):
                     pbr[k] = v
                 elif "index" in v:
                     # get the index of image for texture
-                    idx = header["textures"][v["index"]]["source"]
-                    # store the actual image as the value
-                    pbr[k] = images[idx]
+                    try:
+                        idx = header["textures"][v["index"]]["source"]
+                        # store the actual image as the value
+                        pbr[k] = images[idx]
+                    except BaseException:
+                        log.debug('unable to store texture',
+                                  exc_info=True)
             # create a PBR material object for the GLTF material
             materials.append(visual.material.PBRMaterial(**pbr))
 
     return materials
 
 
-def unique_name(start, contains):
-    """
-    Deterministically generate a unique name not
-    contained in a dict. Will create names of the
-    form "start_10" and increment accordingly.
-
-    Parameters
-    -----------
-    start : str
-      Initial guess for name
-    contains : dict, set, or list
-      Bundle of existing names we cannot use.
-
-    Returns
-    ---------
-    unique : str
-      A name that is not contained in `contains`
-    """
-    # exit early if name is not in bundle
-    if len(contains) == 0 or (len(start) > 0 and start not in contains):
-        return start
-
-    increment = 0
-    formatter = start + '_{}'
-    if len(start) > 0:
-        # split by our delimiter once
-        split = start.rsplit('_', 1)
-        if len(split) == 2 and split[1].isnumeric():
-            if split[0] not in contains:
-                return split[0]
-            # start incrementing from the passed value
-            increment = int(split[1])
-            formatter = split[0] + '_{}'
-
-    # if contains is empty we will only need to check once
-    for i in range(increment + 1, 2 + increment + len(contains)):
-        check = formatter.format(i)
-        if check not in contains:
-            return check
-
-    raise ValueError('unable to establish unique name!')
-
-
 def _read_buffers(header,
                   buffers,
                   mesh_kwargs,
@@ -1296,7 +1342,6 @@ def _read_buffers(header,
                 start = 0
             end = start + view["byteLength"]
             views[i] = buffers[view["buffer"]][start:end]
-
             assert len(views[i]) == view["byteLength"]
 
         # load data from buffers into numpy arrays
@@ -1308,12 +1353,13 @@ def _read_buffers(header,
             # what is the datatype
             dtype = np.dtype(_dtypes[a["componentType"]])
             # basically how many columns
+            # for types like (4, 4)
             per_item = _shapes[a["type"]]
             # use reported count to generate shape
             shape = np.append(count, per_item)
             # number of items when flattened
             # i.e. a (4, 4) MAT4 has 16
-            per_count = np.abs(np.product(per_item))
+            per_count = np.abs(np.prod(per_item))
             if 'bufferView' in a:
                 # data was stored in a buffer view so get raw bytes
 
@@ -1336,7 +1382,7 @@ def _read_buffers(header,
                     # the total block we're looking at
                     length = count * stride
                     # we want to get the bytes for every row
-                    per_row = per_item * dtype.itemsize
+                    per_row = per_count * dtype.itemsize
                     # we have to offset the (already offset) buffer
                     # and then pull chunks per-stride
                     # do as a list comprehension as the numpy
@@ -1365,33 +1411,36 @@ def _read_buffers(header,
     # load data from accessors into Trimesh objects
     meshes = collections.OrderedDict()
 
-    names_original = collections.defaultdict(list)
-
+    # keep track of how many times each name has been attempted to
+    # be inserted to avoid a potentially slow search through our
+    # dict of names
+    name_counts = {}
     for index, m in enumerate(header.get("meshes", [])):
-
         try:
             # GLTF spec indicates implicit units are meters
             metadata = {'units': 'meters'}
             # try to load all mesh metadata
             if isinstance(m.get('extras'), dict):
                 metadata.update(m['extras'])
+            # put any mesh extensions in a field of the metadata
+            if 'extensions' in m:
+                metadata['gltf_extensions'] = m['extensions']
 
-            for j, p in enumerate(m["primitives"]):
+            for p in m["primitives"]:
                 # if we don't have a triangular mesh continue
                 # if not specified assume it is a mesh
                 kwargs = {"metadata": {}, "process": False}
                 kwargs.update(mesh_kwargs)
                 kwargs["metadata"].update(metadata)
-
                 # i.e. GL_LINES, GL_TRIANGLES, etc
-                mode = p.get('mode')
+                # specification says the default mode is GL_TRIANGLES
+                mode = p.get('mode', _GL_TRIANGLES)
                 # colors, normals, etc
                 attr = p['attributes']
                 # create a unique mesh name per- primitive
                 name = m.get('name', 'GLTF')
-                names_original[index].append(name)
                 # make name unique across multiple meshes
-                name = unique_name(name, meshes)
+                name = unique_name(name, meshes, counts=name_counts)
 
                 if mode == _GL_LINES:
                     # load GL_LINES into a Path object
@@ -1401,28 +1450,26 @@ def _read_buffers(header,
                         points=np.arange(len(kwargs['vertices'])))]
                 elif mode == _GL_POINTS:
                     kwargs["vertices"] = access[attr["POSITION"]]
-                elif mode is None or mode in (_GL_TRIANGLES, _GL_STRIP):
-                    if mode is None:
-                        # some people skip mode since GL_TRIANGLES
-                        # is apparently the de-facto default
-                        log.debug(
-                            'primitive has no mode! trying GL_TRIANGLES?')
-                        # get vertices from accessors
+                elif mode in (_GL_TRIANGLES, _GL_STRIP):
+                    # get vertices from accessors
                     kwargs["vertices"] = access[attr["POSITION"]]
                     # get faces from accessors
                     if 'indices' in p:
                         if mode == _GL_STRIP:
                             # this is triangle strips
                             flat = access[p['indices']].reshape(-1)
-                            kwargs['faces'] = util.triangle_strips_to_faces([flat])
+                            kwargs['faces'] = util.triangle_strips_to_faces(
+                                [flat])
                         else:
-                            kwargs["faces"] = access[p["indices"]].reshape((-1, 3))
+                            kwargs["faces"] = access[p["indices"]
+                                                     ].reshape((-1, 3))
                     else:
                         # indices are apparently optional and we are supposed to
                         # do the same thing as webGL drawArrays?
                         kwargs['faces'] = np.arange(
-                            len(kwargs['vertices']),
+                            len(kwargs['vertices']) * 3,
                             dtype=np.int64).reshape((-1, 3))
+
                     if 'NORMAL' in attr:
                         # vertex normals are specified
                         kwargs['vertex_normals'] = access[attr['NORMAL']]
@@ -1430,7 +1477,7 @@ def _read_buffers(header,
                     visuals = None
                     if "material" in p:
                         if materials is None:
-                            log.warning('no materials! `pip install pillow`')
+                            log.debug('no materials! `pip install pillow`')
                         else:
                             uv = None
                             if "TEXCOORD_0" in attr:
@@ -1451,7 +1498,8 @@ def _read_buffers(header,
                                     # just pass to mesh as vertex color
                                     kwargs['vertex_colors'] = colors
                                 else:
-                                    # we ALSO have texture so save as vertex attribute
+                                    # we ALSO have texture so save as vertex
+                                    # attribute
                                     visuals.vertex_attributes['color'] = colors
                         except BaseException:
                             # survive failed colors
@@ -1464,7 +1512,6 @@ def _read_buffers(header,
                     # each primitive gets it's own Trimesh object
                     if len(m["primitives"]) > 1:
                         kwargs['metadata']['from_gltf_primitive'] = True
-                        name = unique_name(name, meshes)
                     else:
                         kwargs['metadata']['from_gltf_primitive'] = False
 
@@ -1474,7 +1521,7 @@ def _read_buffers(header,
                     if len(custom) > 0:
                         kwargs["vertex_attributes"] = custom
                 else:
-                    log.warning('skipping primitive with mode %s!', mode)
+                    log.debug('skipping primitive with mode %s!', mode)
                     continue
                 # this should absolutely not be stomping on itself
                 assert name not in meshes
@@ -1539,16 +1586,20 @@ def _read_buffers(header,
         [meshes.pop(p, None) for p in mesh_pop]
 
     # make it easier to reference nodes
-    nodes = header["nodes"]
+    nodes = header.get("nodes", [])
     # nodes are referenced by index
     # save their string names if they have one
     # we have to accumulate in a for loop opposed
     # to a dict comprehension as it will be checking
     # the mutated dict in every loop
     name_index = {}
+    name_counts = {}
     for i, n in enumerate(nodes):
         name_index[unique_name(
-            n.get('name', str(i)), name_index)] = i
+            n.get('name', str(i)),
+            name_index,
+            counts=name_counts
+        )] = i
     # invert the dict so we can look up by index
     # node index (int) : name (str)
     names = {v: k for k, v in name_index.items()}
@@ -1571,23 +1622,35 @@ def _read_buffers(header,
         # otherwise just use the first index
         scene_index = 0
 
-    # start the traversal from the base frame to the roots
-    for root in header["scenes"][scene_index]["nodes"]:
-        # add transform from base frame to these root nodes
-        queue.append([base_frame, root])
+    if 'scenes' in header:
+        # start the traversal from the base frame to the roots
+        for root in header["scenes"][scene_index].get("nodes", []):
+            # add transform from base frame to these root nodes
+            queue.append((base_frame, root))
+
+    # make sure we don't process an edge multiple times
+    consumed = set()
 
     # go through the nodes tree to populate
     # kwargs for scene graph loader
     while len(queue) > 0:
         # (int, int) pair of node indexes
-        a, b = queue.pop()
+        edge = queue.pop()
+
+        # avoid looping forever if someone specified
+        # recursive nodes
+        if edge in consumed:
+            continue
+
+        consumed.add(edge)
+        a, b = edge
 
         # dict of child node
         # parent = nodes[a]
         child = nodes[b]
         # add edges of children to be processed
         if "children" in child:
-            queue.extend([[b, i] for i in child["children"]])
+            queue.extend([(b, i) for i in child["children"]])
 
         # kwargs to be passed to scene.graph.update
         kwargs = {"frame_from": names[a], "frame_to": names[b]}
@@ -1597,10 +1660,11 @@ def _read_buffers(header,
         # for the transform from parent to child
         if "matrix" in child:
             kwargs["matrix"] = np.array(
-                child["matrix"], dtype=np.float64).reshape((4, 4)).T
+                child["matrix"],
+                dtype=np.float64).reshape((4, 4)).T
         else:
             # if no matrix set identity
-            kwargs["matrix"] = np.eye(4)
+            kwargs["matrix"] = _EYE
 
         # Now apply keyword translations
         # GLTF applies these in order: T * R * S
@@ -1621,9 +1685,16 @@ def _read_buffers(header,
                 kwargs["matrix"],
                 np.diag(np.concatenate((child['scale'], [1.0]))))
 
-        if "extras" in child:
+        # treat node metadata similarly to mesh metadata
+        if isinstance(child.get("extras"), dict):
             kwargs["metadata"] = child["extras"]
 
+        # put any node extensions in a field of the metadata
+        if "extensions" in child:
+            if "metadata" not in kwargs:
+                kwargs["metadata"] = {}
+            kwargs["metadata"]["gltf_extensions"] = child["extensions"]
+
         if "mesh" in child:
             geometries = mesh_prim[child["mesh"]]
 
@@ -1632,11 +1703,11 @@ def _read_buffers(header,
                 # append root node
                 graph.append(kwargs.copy())
                 # put primitives as children
-                for i, geom_name in enumerate(geometries):
+                for geom_name in geometries:
                     # save the name of the geometry
                     kwargs["geometry"] = geom_name
                     # no transformations
-                    kwargs["matrix"] = np.eye(4)
+                    kwargs["matrix"] = _EYE
                     kwargs['frame_from'] = names[b]
                     # if we have more than one primitive assign a new UUID
                     # frame name for the primitives after the first one
@@ -1666,6 +1737,15 @@ def _read_buffers(header,
             header['scene']]['extras']
     except BaseException:
         pass
+    try:
+        # load any scene extensions into a field of scene.metadata
+        # use a try except to avoid nested key checks
+        if "metadata" not in result:
+            result["metadata"] = {}
+        result['metadata']['gltf_extensions'] = header['scenes'][
+            header['scene']]['extensions']
+    except BaseException:
+        pass
 
     return result
 
